# Generated by Django 4.1.2 on 2023-02-17 23:34

import demo.users.models
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('users', '0003_api_group'),
    ]

    operations = [
        migrations.CreateModel(
            name='APIUser',
            fields=[
            ],
            options={
                'verbose_name': 'API User',
                'verbose_name_plural': 'API Users',
                'proxy': True,
                'indexes': [],
                'constraints': [],
            },
            bases=('users.user',),
            managers=[
                ('objects', demo.users.models.APIUserManager()),
            ],
        ),
    ]
