{% extends 'academy_theme/base--wildewidgets.html' %}
{% load static sphinx_hosting %}

{% block extra_css %}
{{ block.super }}
<link rel="stylesheet" href="{% static 'sphinx_hosting/css/sphinx_hosting.css' %}">
{% endblock %}

{% block extra_footer_js %}
{{ block.super }}
<script src="https://cdnjs.cloudflare.com/ajax/libs/fslightbox/3.4.1/index.js" integrity="sha512-jrYR1cG7wwq2l+uNH+XXF18hjN+j8MBjM2PK2+fV/nAIHKxqpg479rOWFOxTpCnyPMZeGAi+eDAxHyrzTzkyRg==" crossorigin="anonymous" referrerpolicy="no-referrer"></script>
<script async="async" src="https://cdnjs.cloudflare.com/ajax/libs/mathjax/3.2.2/es5/tex-mml-chtml.min.js" integrity="sha512-6FaAxxHuKuzaGHWnV00ftWqP3luSBRSopnNAA2RvQH1fOfnF/A1wOfiUWF7cLIOFcfb1dEhXwo5VG3DAisocRw==" crossorigin="anonymous" referrerpolicy="no-referrer"></script>
<script>
    fsLightbox.props.type = "image";
</script>
{% endblock %}