# Generated by Django 5.1.4 on 2024-12-17 17:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('sphinxhostingcore', '0016_project_last_version_alter'),
    ]

    operations = [
        migrations.AlterField(
            model_name='sphinxpage',
            name='body',
            field=models.TextField(blank=True, help_text='The body for the page, extracted from the page JSON, and modified to suit us. Some pages have no body.  The body is actually stored as a Django template.', verbose_name='Body'),
        ),
        migrations.AlterField(
            model_name='sphinxpage',
            name='orig_global_toc',
            field=models.TextField(blank=True, default=None, help_text='The original global table of contents HTML attached to this page, if any.  This will only be present if you had "sphinxcontrib-jsonglobaltoc"installed in your "extensions" in the Sphinx conf.py', null=True, verbose_name='Global Table of Contents (original)'),
        ),
    ]
